use std::path::{Path, PathBuf};
use reqwest::{
    header::{HeaderMap, HeaderName, HeaderValue},
    Method as ReqMethod,
};
use rocket::{
    response::stream::ByteStream,
    data::ToByteUnit,
    http::{Header, Method},
    response::{Redirect},
    route::{Handler, Outcome},
    Data, Request, Route,
};

#[derive(Clone)]
pub struct ProxyServer
{
    root: String,
    prefix: PathBuf,
    rank: isize,
}

impl ProxyServer
{
    const DEFAULT_RANK: isize = 0;

    pub fn new<P: AsRef<Path>>(root: &str, prefix: P, rank: isize) -> Self
    {
        ProxyServer { root: String::from(root), prefix: prefix.as_ref().into(), rank }
    }

    pub fn from<P: AsRef<Path>>(root: &str, prefix: P) -> Self
    {
        Self::new(root, prefix, Self::DEFAULT_RANK)
    }
}

impl Into<Vec<Route>> for ProxyServer
{
    fn into(self) -> Vec<Route>
    {
        let mut routes: Vec<Route> = Vec::new();

        static METHODS: [Method; 5] = [
            Method::Get,
            Method::Post,
            Method::Patch,
            Method::Delete,
            Method::Put,
        ];

        for method in METHODS {
            let mut route = Route::ranked(self.rank, method, "/<path..>", self.clone());
            route.name =
                Some(format!("ProxyServer: {} {}{}", method.as_str(), self.root.clone(), self.prefix.to_str().unwrap()).into());

            routes.push(route);
        }

        routes
    }
}

#[rocket::async_trait]
impl Handler for ProxyServer
{
    async fn handle<'r>(&self, req: &'r Request<'_>, data: Data<'r>) -> Outcome<'r>
    {
        println!("{:?}", req);
        // The goals here is to convert a Rocket request into a reqwest one

        // First, we convert the headers
        let mut headers = HeaderMap::new();

        for header in req.headers().iter() {
            // TODO remove unwrap
            headers.insert(
                HeaderName::from_lowercase(header.name.into_string().as_bytes()).unwrap(),
                HeaderValue::from_str(&header.value.into_owned()).unwrap(),
            );
        }

        // Then the method
        let method = match req.method() {
            Method::Delete => ReqMethod::DELETE,
            Method::Get => ReqMethod::GET,
            Method::Patch => ReqMethod::PATCH,
            Method::Post => ReqMethod::POST,
            Method::Put => ReqMethod::PUT,
            _ => todo!(),
        };

        // Then the URL
        // We first extract all URL segments starting from the mountpoint
        let segments: PathBuf = req.segments(0..).unwrap();

        let query_part = if let Some(query) = req.uri().query() {
            format!("?{}", query)
        }else{
            String::from("")
        };

        let url = format!(
            "{}{}/{}{}",
            self.root,
            self.prefix.to_str().unwrap(),
            segments.to_str().unwrap(),
            query_part
        );

        // And finally, the data
        // TODO don't hard-code max request size here
        static BODY_SIZE: i64 = 2;
        // TODO remove unwrap
        let data: Vec<u8> = data
            .open(BODY_SIZE.mebibytes())
            .into_bytes()
            .await
            .unwrap()
            .to_vec();

        // TODO remove unwrap
        let new_res = reqwest::Client::new()
            .request(method, url)
            .headers(headers)
            .body(data)
            .send()
            .await
            .unwrap();
        println!("{:?}", new_res);

        let mut res_builder = rocket::Response::build();
        let mut res_builder = res_builder.status(rocket::http::Status::new(new_res.status().as_u16()));
            // .streamed_body(ByteStream::from(new_res.bytes_stream()));

        // reqwest headers -> rocket headers
        for (key, value) in new_res.headers().clone().iter_mut() {
            res_builder = res_builder.raw_header(
                String::from(key.clone().as_str()),
                String::from(value.clone().to_str().unwrap())
            );

        }

        Outcome::Success(res_builder.finalize())
    }
}
